#  tests for geopandas-0.4.0-py36_1 (this is a generated file);
print('===== testing package: geopandas-0.4.0-py36_1 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
import os
import geopandas

fname = os.path.join(os.environ['RECIPE_DIR'], 'test_data', 'test.shp')
geopandas.read_file(fname)
#  --- run_test.py (end) ---

print('===== geopandas-0.4.0-py36_1 OK =====');
print("import: 'geopandas'")
import geopandas

print("import: 'geopandas.io'")
import geopandas.io

